#include "pch.h"
#include "common.h"


static bool checkPlatform()
{
	int sizeOfVstIntPtr = sizeof (VstIntPtr);
	int sizeOfVstInt32 = sizeof (VstInt32);
	int sizeOfPointer = sizeof (void*);
	int sizeOfAEffect = sizeof (AEffect);

	return sizeOfVstIntPtr == sizeOfPointer;
}


//-------------------------------------------------------------------------------------------------------

static void checkEffectProperties (AEffect* effect);
static void checkEffectProcessing (AEffect* effect);
extern bool checkEffectEditor (AEffect* effect); // minieditor.cpp

extern float channelDummyIn[MAX_AUDIO_BUFFER] ;
extern float channelDummyOut[MAX_AUDIO_BUFFER];

float WorkAudioBufferF[2][MAX_AUDIO_BUFFER] = {0};
float WorkAudioBufferOutF[2][MAX_AUDIO_BUFFER] = {0};

extern void getEffectName(PluginData *pd, char *str, size_t sz);

std::vector<PluginData> ChannelPluginList;

extern bool openEffectEditor(PluginData *pd, HWND hwnd, bool createOnly);
extern void effectProcessing_ClearVstBuffer(AEffect *effect);
extern void effectProcessingBase(float **in, float **out, long count, float freq);
//-------------------------------------------------------------------------------------------------------

struct EffectProcessingManager
{
	float *in[MAX_INOUT_CHANNEL];
	float *out[MAX_INOUT_CHANNEL];

	EffectProcessingManager()
	{
		in[0] = WorkAudioBufferF[0];
		in[1] = WorkAudioBufferF[1];
		out[0] = WorkAudioBufferOutF[0];
		out[1] = WorkAudioBufferOutF[1];
		for (int i = 2; i < MAX_INOUT_CHANNEL; ++i) {
			in[i] = channelDummyIn;
			out[i] = channelDummyOut;
		}
	}
};

char *script_spaceskip(char *ptr)
{
	char *p = ptr;
	while (*p == ' ' || *p == '\t')p++;
	return p;
}

char *script_spaceskip_reverse(char *ptr)
{
	char *p = ptr;
	while (*p == ' ' || *p == '\t')p--;
	return p;
}

void PluginList_DeleteEndBlock()
{
	mutex.lock();
	PluginList.erase( PluginList.end()-1 );
	mutex.unlock();
}

void PluginChannelList_DeleteEndBlock()
{
	mutex.lock();
	ChannelPluginList.erase( ChannelPluginList.end()-1 );
	mutex.unlock();

}

extern void WINAPI effectInit();

static bool bFirstInitEnd = false;
char strDllPath[MAX_PATH + 4]="";
char strPresetPath[MAX_PATH + 4]="";

void GetVstCurrentPath(char *str, DWORD size)
{
	strncpy(str, strDllPath, size);
}

bool VST_Loader( PluginData *pld, char *path, void (* PluginListBackDeleter)())
{
	char *p = path;
	pld->hwnd = NULL;
	pld->bShow = false;
	pld->bSystemEnd = false;
	pld->bClearVstBuffer = false;
	pld->bSampleRateInitYes = false;

	char *endchk = strstr(p, "#");
	if (endchk) {
		endchk = script_spaceskip_reverse(--endchk);
		*++endchk = 0;
	}

	char strFullPath[MAX_PATH+4]="";
	if(PathIsRelative(p)) {
		GetVstCurrentPath(strFullPath, MAX_PATH);
		strcat(strFullPath, "\\");
		strcat(strFullPath, p);
	}else strcpy(strFullPath, p);

	char dllAbsPath[MAX_PATH+4]="";
	char timCurrentPath[MAX_PATH+4]="";
	strcpy(dllAbsPath, strFullPath);
	PathRemoveFileSpec(dllAbsPath);
	SetDllDirectory(dllAbsPath);

	bool loaderResult = pld->loader.loadLibrary(strFullPath);
	
	if (!loaderResult) {
		PluginListBackDeleter();
		return false;
	}

	pld->mainEntry = pld->loader.getMainEntry();
	if (!pld->mainEntry){
		PluginListBackDeleter();
		return false;
	}

	pld->effect = pld->mainEntry(HostCallback);
	if (!pld->effect){
		PluginListBackDeleter();
		return false;
	}

	pld->effect->dispatcher(pld->effect, effOpen, 0, 0, 0, 0);

	checkEffectProperties(pld->effect);

	return true;
}



int __stdcall vstOpen()
{
	bFirstInitEnd = false;

	if (!checkPlatform ()){
		return -1;
	}

	IniMemFile ini;
	ini.LoadFromFile("soundfont.ini");
	g_MaxChannelVstCount = ini.ReadInt("VST", "MaxChannel", 8);
	if (g_MaxChannelVstCount < 2)
		g_MaxChannelVstCount = 2;

	if (g_MaxChannelVstCount != 0) {
		if ( ChannelVstItem != NULL ) {
			MessageBox(NULL, "memory address damage ?", "Critical Error", MB_OK);
		}
		mutex.lock();
		ChannelVstItem = new tagChannelVstItem[g_MaxChannelVstCount]();
		for(int i = 0; i<g_MaxChannelVstCount; ++i) {
			ChannelVstItem[i].chMask = 0;
			ChannelVstItem[i].ID_VSTInsertBtn=0;
			ChannelVstItem[i].ID_VSTInsertBtn_VstLoad=0;
			strcpy(ChannelVstItem[i].name, "");
		}

		mutex.unlock();
	}else{

	}

	const int buff_size = 4096;
	char buff[buff_size + 4];
	bool TrackChannelLoaderOn = false;

	char strVstListPath[MAX_PATH+4]="";
	GetVstCurrentPath(strVstListPath, MAX_PATH);
	strcat(strVstListPath, "\\vst_list.txt");
	FILE *fp = NULL;
	{
		char strVstListCachePath[MAX_PATH+4]="";
		GetVstCurrentPath(strVstListCachePath, MAX_PATH);
		strcat(strVstListCachePath, "\\preset\\vst_list.cache.txt");
		fp = fopen(strVstListCachePath, "r");
		if (fp != NULL) {
			FILETIME ctime, atime, utime, chk_time;
			fscanf(fp, "%x,%x", &chk_time.dwHighDateTime, &chk_time.dwLowDateTime);

			HANDLE hFile = CreateFile(strVstListPath, GENERIC_READ, 0, NULL,OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
			if ( hFile == INVALID_HANDLE_VALUE) {

			}else{
				GetFileTime(hFile, &ctime, &atime, &utime);
				CloseHandle(hFile);

				if (utime.dwHighDateTime == chk_time.dwHighDateTime && utime.dwLowDateTime == chk_time.dwLowDateTime) {
					
				}else{
					fclose(fp);
					fp = fopen(strVstListPath, "r");
				}
			}
		}else{
			fp = fopen(strVstListPath, "r");
		}
	}
	if (fp == NULL)return 1;



	while ( fgets(buff, buff_size, fp) != NULL ) {
		char *p = strstr(buff, "\n");
		if (p)*p = 0;

		p = script_spaceskip(buff);
		if (*p == '#')continue;
		if (*p == '\0')continue;
		if (strncmp(p, "__END__", 7) == 0)break;
		if (strncmp(p, "__TRACK__", 9) == 0)TrackChannelLoaderOn = true;
		if (strncmp(p, "__TRACK_END__", 13) == 0)TrackChannelLoaderOn = false;

		char *p2 = strstr(p, "$");
		if (p2)*p2++ = 0;

		if (TrackChannelLoaderOn) {
			mutex.lock();
			ChannelPluginList.push_back(PluginData());
			PluginData *pld = &ChannelPluginList[ PluginList.size() - 1 ];
			pld->index = PluginList.size() - 1;
			VST_Loader(pld, p, PluginChannelList_DeleteEndBlock);
			mutex.unlock();
		}else{
			mutex.lock();
			PluginList.push_back(PluginData());
			PluginData *pld = &PluginList[ PluginList.size() - 1 ];
			pld->index = PluginList.size() - 1;
			mutex.unlock();

			if (p2)pld->bProsessing = (*p2 == '0') ? false : true;
			else pld->bProsessing = true;

			mutex.lock();
			VST_Loader(pld, p, PluginList_DeleteEndBlock);
			mutex.unlock();
		}
	}

	fclose(fp);

	for (int i = 0; i < (int)PluginList.size();  ++i) {
		openEffectEditor(&PluginList[i], NULL, true);
	}

	Sleep(100);

	bFirstInitEnd = true;

	return TRUE;
}

extern HWND hwndManager;

void vstCloseLastSave()
{
	char strVstListCachePath[MAX_PATH+4]="";
	GetVstCurrentPath(strVstListCachePath, MAX_PATH);
	strcat(strVstListCachePath, "\\preset\\vst_list.cache.txt");
	FILE *fp = fopen(strVstListCachePath, "w");

	if (fp == NULL) { return; }

	char strVstListPath[MAX_PATH+4]="";
	GetVstCurrentPath(strVstListPath, MAX_PATH);
	strcat(strVstListPath, "\\vst_list.txt");

	FILETIME ctime, atime, utime;
	HANDLE hFile = CreateFile(strVstListPath, GENERIC_READ, 0, NULL,OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	GetFileTime(hFile, &ctime, &atime, &utime);
	CloseHandle(hFile);
	fprintf(fp, "%x,%x\n", utime.dwHighDateTime, utime.dwLowDateTime);


	for (size_t i = 0; i < PluginList.size(); ++i) {
		char fn[MAX_PATH+4]="";
		GetModuleFileName(PluginList[i].loader.module, fn, MAX_PATH);
		fprintf(fp, "%s$%d\n", fn, (int)PluginList[i].bProsessing);
	}
	fprintf(fp, "__END__\n");
	fclose(fp);
	fp = NULL;
}

void vstAllFree()
{
	for (size_t i = 0; i < PluginList.size(); ++i) {
		PluginList[i].bProsessing = PluginList[i].bSystemEnd = true;
		SendMessage(PluginList[i].hwnd, WM_CLOSE, 0, 0);
	}
	Sleep(1000);
	while (g_processingLock.isLocked()) {Sleep(10);}

	mutex.lock();
	for (size_t i = 0; i < PluginList.size(); ++i) {
		PluginList[i].effClose();
		PluginList[i].loader.freeLibrary();
	}
	mutex.unlock();

	while (g_processingLock.isLocked()) {Sleep(10);}
	Sleep(10);

	mutex.lock();
	PluginList.clear();
	mutex.unlock();


	mutex.lock();
	for (int i = 0; i < g_MaxChannelVstCount; ++i) {
		ChannelVstItem[i].pd.clear();
	}
	mutex.unlock();

	if (hwndManager) {
		SendMessage(hwndManager, WM_CLOSE, 0, 0);
	}

	mutex.lock();
	delete[] ChannelVstItem;
	ChannelVstItem = NULL;
	mutex.unlock();
}

void __stdcall vstClose()
{
	vstCloseLastSave();
	vstAllFree();
}

void __stdcall openEffectEditorAll(HWND hwnd)
{
	for (size_t i = 0; i < PluginList.size(); ++i) {
		openEffectEditor(&PluginList[i], hwnd, false);
	}
}

struct tagPresetFooter
{
	unsigned long hdr;
	int count;
	unsigned long eof;
};

struct tagChunkFooter
{
	unsigned long hdr;
	unsigned long eof;
};

void loadPlugin_CustomPreset(const char *filename, AEffect *effect)
{
	FILE *fp = fopen(filename, "rb");
	if (fp==NULL)return;

	fseek(fp, -(int)(sizeof tagPresetFooter), SEEK_END);
	tagPresetFooter footer = {0};
	fread(&footer, sizeof tagPresetFooter, 1, fp);
	fseek(fp, 0, SEEK_SET);

	//check.
	{
		int err_check = 0;
		if (footer.hdr != *(DWORD *)"VPS1") 
			err_check++;
		if (footer.eof != *(DWORD *)"EOF ")
			err_check++;
		if (footer.count != effect->numParams)
			err_check++;
		if (err_check) {
			fclose(fp);
			return;
		}
	}

	float f = 0.0f;
	for (VstInt32 i = 0; i < footer.count; i++){
		fread(&f, sizeof(float), 1, fp);
		effect->setParameter(effect, i, f);
	}

	fclose(fp);
}

void loadChunk(PluginData *pld, const char *fn)
{
	AEffect *effect = pld->effect;

	if (effect->flags & effFlagsProgramChunks) {

		int bIsBank = false;

		try {
			if (stricmp(PathFindExtension(fn), ".fxb") == 0) {
				bIsBank = 0;
			}else if (stricmp(PathFindExtension(fn), ".fxp") == 0) {
				bIsBank = 1;
			}

			char *data = NULL;
			FILE *fp = NULL;
			int sz = 0;

			fp = fopen(fn, "rb");
			if (fp == NULL)return;
			fread(&sz, sizeof(int), 1, fp);
			if (!sz) { fclose(fp); return; }
			
			mutex.lock();
			data = (char *)malloc(sz);
			mutex.unlock();

			fread(data, 1, sz, fp);
			pld->effect->dispatcher(pld->effect, effSetChunk, bIsBank, sz, data, 0 );

			mutex.lock();
			free(data);
			mutex.unlock();

			fclose(fp);

		}catch(...){
			MessageBox(NULL, "Load Chunk Failed", "Error", MB_OK);
		}
	}
}

void loadPluginPreset(PluginData *pld)
{
	char effectName[MAX_PATH + 4] = {0};
	getEffectName(pld, effectName, MAX_PATH);

	char path[MAX_PATH + 4];

	sprintf(path, "%s/pid%d_%s.fxb", strPresetPath, pld->index, effectName);
	if (!PathFileExists(path))
		sprintf(path, "%s/%s.fxb", strPresetPath, effectName);
	loadChunk(pld, path);

	sprintf(path, "%s/pid%d_%s.pri", strPresetPath, pld->index, effectName);
	if (!PathFileExists(path))
			sprintf(path, "%s/%s.pri", strPresetPath, effectName);
	loadPlugin_CustomPreset(path, pld->effect);
}

void savePlugin_CustomPreset(const char *filename, AEffect *effect)
{
	FILE *fp = fopen(filename, "wb");
	if (fp == NULL)return;
	
	float f;
	const VstInt32 numParams = effect->numParams;
	for (VstInt32 i = 0; i < numParams; i++){
		f = effect->getParameter(effect, i);
		fwrite(&f, sizeof(float), 1, fp);
	}

	tagPresetFooter footer = {0};
	footer.count = numParams;
	footer.hdr = *(DWORD *)"VPS1";
	footer.eof = *(DWORD *)"EOF ";
	fwrite(&footer, sizeof tagPresetFooter, 1, fp);

	fclose(fp);
}

void saveChunk(PluginData *pld, const char *fn)
{
	AEffect *effect = pld->effect;

	if (effect->flags & effFlagsProgramChunks) {
		
		FILE *fp = NULL;
		int bIsBank = false;

		try {
			if (stricmp(PathFindExtension(fn), ".fxb") == 0) {
				bIsBank = 0;
			}else if (stricmp(PathFindExtension(fn), ".fxp") == 0) {
				bIsBank = 1;
			}

			int sz = 0;
			char *data = NULL;

			sz = effect->dispatcher(effect, effGetChunk, bIsBank, 0, &data, 0.0f );
			if (sz == 0)return;

			fp = fopen(fn, "wb");
			if (fp == NULL)return;
			fwrite(&sz, sizeof(int), 1, fp);
			fwrite(data, 1, sz, fp);
			fclose(fp);
		}catch(...){
			MessageBox(NULL, "Save Chunk Failed", "Error", MB_OK);
		}
	}
}

void savePluginPreset(PluginData *pld)
{
	AEffect *effect = pld->effect;
	char effectName[MAX_PATH+4] = {0};
	getEffectName(pld, effectName, MAX_PATH);

	char path[MAX_PATH+4]="";

	mutex.lock();
	sprintf(path, "%s/pid%d_%s.fxb", strPresetPath, pld->index, effectName);
	saveChunk(pld, path);
	mutex.unlock();

	mutex.lock();
	sprintf(path, "%s/pid%d_%s.pri", strPresetPath, pld->index, effectName);
	savePlugin_CustomPreset(path, effect);
	mutex.unlock();
}

//-------------------------------------------------------------------------------------------------------
void checkEffectProperties (AEffect* effect)
{
	printf ("HOST> Gathering properties...\n");

	char effectName[256] = {0};
	char vendorString[256] = {0};
	char productString[256] = {0};

//	effect->dispatcher (effect, effGetEffectName, 0, 0, effectName, 0);
//	effect->dispatcher (effect, effGetVendorString, 0, 0, vendorString, 0);
//	effect->dispatcher (effect, effGetProductString, 0, 0, productString, 0);

//	printf("Name = %s\nVendor = %s\nProduct = %s\n\n", effectName, vendorString, productString);

//	printf("numPrograms = %d\nnumParams = %d\nnumInputs = %d\nnumOutputs = %d\n\n", 
//			effect->numPrograms, effect->numParams, effect->numInputs, effect->numOutputs);
#if 0
	// Iterate programs...
	for (VstInt32 progIndex = 0; progIndex < effect->numPrograms; progIndex++)
	{
		char progName[256] = {0};
		if (!effect->dispatcher(effect, effGetProgramNameIndexed, progIndex, 0, progName, 0))
		{
			effect->dispatcher(effect, effSetProgram, 0, progIndex, 0, 0); // Note: old program not restored here!
			effect->dispatcher(effect, effGetProgramName, 0, 0, progName, 0);
		}
		printf("Program %03d: %s\n", progIndex, progName);
	}

	// Iterate parameters...
	for (VstInt32 paramIndex = 0; paramIndex < effect->numParams; paramIndex++)
	{
		char paramName[256] = {0};
		char paramLabel[256] = {0};
		char paramDisplay[256] = {0};

		effect->dispatcher(effect, effGetParamName, paramIndex, 0, paramName, 0);
		effect->dispatcher(effect, effGetParamLabel, paramIndex, 0, paramLabel, 0);
		effect->dispatcher(effect, effGetParamDisplay, paramIndex, 0, paramDisplay, 0);
		float value = effect->getParameter(effect, paramIndex);

		printf("Param %03d: %s [%s %s] (normalized = %f)\n", paramIndex, paramName, paramDisplay, paramLabel, value);
	}
	printf ("\n");
#endif
	
	// Can-do nonsense...
	static const char* canDos[] =
	{
		"receiveVstEvents",
		"receiveVstMidiEvent",
		"midiProgramNames"
	};

	for (VstInt32 canDoIndex = 0; canDoIndex < sizeof (canDos) / sizeof (canDos[0]); canDoIndex++)
	{
		printf ("Can do %s... ", canDos[canDoIndex]);
		VstInt32 result = (VstInt32)effect->dispatcher(effect, effCanDo, 0, 0, (void*)canDos[canDoIndex], 0);
		switch (result)
		{
			case 0  : printf("don't know"); break;
			case 1  : printf("yes"); break;
			case -1 : printf("definitely not!"); break;
			default : printf("?????");
		}
		printf ("\n");
	}

	printf ("\n");
}

//-------------------------------------------------------------------------------------------------------

#pragma warning(push)
#pragma warning(disable: 4244)

class WindowFunction
{
	enum {
		MAX_LEN = 1024,
	};

	float win_[MAX_LEN];
	bool bInit_;
public:
	WindowFunction()
	{
		bInit_ = false;
	}

	void InitHanning()
	{
		long length = MAX_LEN;
		double pi, phase = 0, delta;
		pi = 4.*atan(1.0);
    	delta = 2 * pi / (double) length;
    	for (int i=0;i<length;i++)	{
			win_[i] = (float) (0.5 * (1.0 - cos(phase)));
			phase += delta;
    	}
	}

	void Do(float *p, int len)
	{
		if (!bInit_) {
			InitHanning();
			bInit_ = true;
		}
		for ( int i = 0; i < len; ++i ) {
			p[i] *= win_[i % MAX_LEN];
		}
	}

};

WindowFunction g_wf;


inline long Sample_ConvFtoI32(float v)
{
	if (v == 0.0f)return 0;
	if (v < 0.0)
		return - (v * (float)LONG_MIN);
	return v * (float)LONG_MAX;
}

inline float Sample_ConvI32toF(long v)
{
	if (v == 0)return 0.0f;
	if (v < 0.0)
		return - ((float)(v) / (float)LONG_MIN);
	return (float)v / (float)LONG_MAX;
}

inline short Sample_ConvFtoI16(float v)
{
	long n = 0;
	if (v < 0.0) 
		n = - (long)(v * (float)SHRT_MIN + 0.5f);
	else
		n = (long)(v * (float)SHRT_MAX + 0.5f);
	if (n < SHRT_MIN)return SHRT_MIN;
	if (n > SHRT_MAX)return SHRT_MAX;
	return n;

}

inline float Sample_ConvI16toF(short v)
{
	if (v == 0)return 0.0f;
	if (v < 0.0)
		return - ((float)(v-0.5) / (float)SHRT_MIN);
	return (float)(v+0.5) / (float)SHRT_MAX;
}

inline char Sample_ConvFtoI8(float v)
{
	if (v == 0.0f)return 0;
	if (v < 0.0)
		return - (v * (float)CHAR_MIN);
	return v * (float)CHAR_MAX;
}

inline float Sample_ConvI8toF(char v)
{
	if (v == 0)return 0.0f;
	if (v < 0.0)
		return - ((float)v / (float)CHAR_MIN);
	return (float)v / (float)CHAR_MAX;
}

inline float Sample_Int24toFloat(unsigned char *buff)
{
	unsigned long n=0;
	reinterpret_cast<unsigned char *>(&n)[2] = buff[2];
	reinterpret_cast<unsigned char *>(&n)[1] = buff[1];
	reinterpret_cast<unsigned char *>(&n)[0] = buff[0];

	if (n > 0x7FFFFF) // -
		return - (float)((n ^ 0xFFFFFF) + 1) / (float)0x800000; 
	return (float)n / (float)0x7FFFFF;
}

inline long Sample_FloatToInt24(float f)
{
	long n = 0;
	if (f < 0.0f) { // -
		n = ((f * (float)0x800000) - 0.5f); 
		if (n < (-0x800000))n = (-0x800000);
		return n;
	}
	n = f * (float)0x7FFFFF + 0.5f;
	if (n > 0x7FFFFF)n = 0x7FFFFF;
	return n;
}

void sint_32to24(unsigned long v, unsigned char *buff)
{
	buff[2] = reinterpret_cast<unsigned char *>(&v)[2];
	buff[1] = reinterpret_cast<unsigned char *>(&v)[1];
	buff[0] = reinterpret_cast<unsigned char *>(&v)[0];
}

#pragma warning(pop)

void effectProcessing_ClearVstBuffer(AEffect *effect)
{
	const int MAX_CLEAR_BUFF_SIZE = 8192;
	float cin[MAX_CLEAR_BUFF_SIZE * 2] = {0};
	float cout[MAX_CLEAR_BUFF_SIZE * 2] = {0};
	memset(cin, 0, sizeof(cin));
	memset(cout, 0, sizeof(cout));

	float *pcin[MAX_INOUT_CHANNEL]  = {NULL};
	float *pcout[MAX_INOUT_CHANNEL] = {NULL};

	for (int i = 0; i < MAX_INOUT_CHANNEL; ++i) {
		pcin[i] = cin;
		pcout[i] = cout;
	}

	const int loopCount = 48000 / MAX_CLEAR_BUFF_SIZE * 5000 ;
	for (int i = 0; i < loopCount; ++i) {
		mutex.lock();
		memset(cin, 0, sizeof(cin));
		memset(cout, 0, sizeof(cout));
		effect->processReplacing(effect, pcin, pcout, MAX_CLEAR_BUFF_SIZE);
		mutex.unlock();
		{
			bool bLoopSet = false;
			for (int d = 0; d < MAX_CLEAR_BUFF_SIZE; ++d) {
				if (abs(cout[d]) > 0.0000001f) {
					bLoopSet = true;
					break;
				}
			}
			if (!bLoopSet)break;
		}
	}
}

void effectProcessingBase(float **in, float **out, long count, float freq)
{
	if ( ! bFirstInitEnd)
		return;

	static float svFreq = 0;
	int list_length = 0;
	
	if (freq && svFreq != freq) {
		list_length = (int)PluginList.size();
		VSTProcessingWaitLocker lk(g_processingLock);
		svFreq = freq;
		for (int i = 0; i < list_length; ++i) {
			AEffect *effect = PluginList[i].effect;
			if ( ! PluginList[i].bSampleRateInitYes) { // init !
				mutex.lock();
				effect->dispatcher(effect, effSetSampleRate, 0, 0, 0, svFreq);
				effect->dispatcher(effect, effSetBlockSize, 0, MAX_AUDIO_BUFFER, 0, 0);
				effect->dispatcher(effect, effMainsChanged, 0, 1, 0, 0);
				
				PluginList[i].bSampleRateInitYes = true;
				mutex.unlock();
			}else{
				mutex.lock();
				effect->dispatcher(effect, effSetSampleRate, 0, 0, 0, svFreq);
				mutex.unlock();
			}
		}
	}

	const long countbyte = ( sizeof(float) * (count) );
	list_length = (int)PluginList.size();
	VSTProcessingWaitLocker lk(g_processingLock);
	for (int i = 0; i < list_length; ++i) {
		AEffect *effect = PluginList[i].effect;

		if ( ! PluginList[i].bSampleRateInitYes){
			mutex.lock();
			effect->dispatcher(effect, effSetSampleRate, 0, 0, 0, svFreq);
			effect->dispatcher(effect, effSetBlockSize, 0, MAX_AUDIO_BUFFER, 0, 0);
			effect->dispatcher(effect, effMainsChanged, 0, 1, 0, 0);
			
			PluginList[i].bSampleRateInitYes = true;
			mutex.unlock();
		}

		if (PluginList[i].bClearVstBuffer) {// buffer init!
			effectProcessing_ClearVstBuffer(effect);
			PluginList[i].bClearVstBuffer = false;
		}

		if (PluginList[i].bProsessing) {

			if (effect->numInputs == 1) { // mono vst !
				for (int ib = 0; ib < count; ++ib) {
					in[1][ib] = in[0][ib] = (in[0][ib] + in[1][ib]) / 2.0f;
				}
			}

			mutex.lock();
			effect->processReplacing(effect, in, out, count);
			mutex.unlock();

			if (effect->numOutputs == 1) { // mono vst !
				memcpy(in[0], out[0], countbyte);
				memcpy(in[1], out[0], countbyte);
			}else{
				memcpy(in[0], out[0], countbyte);
				memcpy(in[1], out[1], countbyte);
			}
		}else{}
	}
}

void __stdcall effectProcessingInt32(long *buff, long count, float freq)
{
	if (count < 0)return;

	EffectProcessingManager epm;

	for (int i = 0, k = 0; i < count; ++i, ++k) {// convert i32 -> f32
		epm.out[0][k] = epm.in[0][k] = Sample_ConvI32toF(buff[i]);
		++i;
		epm.out[1][k] = epm.in[1][k] = Sample_ConvI32toF(buff[i]);
	}

	effectProcessingBase(epm.in, epm.out, count / 2, freq);

	for (int i = 0, k = 0; i < count; ++i,++k) {// convert f32 -> i32
		buff[i] = Sample_ConvFtoI32(epm.out[0][k]);
		++i;
		buff[i] = Sample_ConvFtoI32(epm.out[1][k]);
	}
}

void __stdcall effectProcessingInt16(short *buff, long count, float freq)
{
	if (count < 0)return;

	EffectProcessingManager epm;

	for (int i = 0, k = 0; i < count; ++i, ++k) {
		epm.out[0][k] = epm.in[0][k] = Sample_ConvI16toF(buff[i]);
		++i;
		epm.out[1][k] = epm.in[1][k] = Sample_ConvI16toF(buff[i]);
	}

	effectProcessingBase(epm.in, epm.out, count / 2, freq);

	for (int i = 0, k = 0; i < count; ++i, ++k) {
		buff[i] = Sample_ConvFtoI16(epm.out[0][k]);
		++i;
		buff[i] = Sample_ConvFtoI16(epm.out[1][k]);
	}
}

void __stdcall effectProcessingInt24(char *buff, long count, float freq)
{
	if (count < 0)return;

	EffectProcessingManager epm;
	unsigned char *p;

	count = count>>1;
	
	p = (unsigned char *)buff + 0;
	for (int k = 0; k < count; ++k) {
		epm.out[0][k] = epm.in[0][k] = Sample_Int24toFloat(p); p+=3;
		epm.out[1][k] = epm.in[1][k] = Sample_Int24toFloat(p); p+=3;
	}

	effectProcessingBase(epm.in, epm.out, count, freq);

	p = (unsigned char *)buff + 0;
	for (int k = 0; k < count; ++k) {
		sint_32to24(Sample_FloatToInt24(epm.out[0][k]), p); p+=3;
		sint_32to24(Sample_FloatToInt24(epm.out[1][k]), p); p+=3;
	}
}

void __stdcall effectProcessingInt8(char *buff, long count, float freq)
{
	if (count < 0)return;

	EffectProcessingManager epm;

	for (int i = 0, k = 0; i < count; ++i, ++k) {
		epm.out[0][k] = epm.in[0][k] = Sample_ConvI8toF(buff[i]);
		++i;
		epm.out[1][k] = epm.in[1][k] = Sample_ConvI8toF(buff[i]);
	}

	effectProcessingBase(epm.in, epm.out, count / 2, freq);

	for (int i = 0, k = 0; i < count; ++i, ++k) {
		buff[i] = Sample_ConvFtoI8(epm.out[0][k]);
		++i;
		buff[i] = Sample_ConvFtoI8(epm.out[1][k]);
	}
}

void __stdcall effectProcessingFloat(float *buff, long count, float freq)
{
	EffectProcessingManager epm;

	for (int i = 0, k = 0; i < count; ++i, ++k) {
		epm.out[0][k] = epm.in[0][k] = buff[i];
		++i;
		epm.out[1][k] = epm.in[1][k] = buff[i];
	}

	effectProcessingBase(epm.in, epm.out, count / 2, freq);

	for (int i = 0, k = 0; i < count; ++i, ++k) {
		buff[i] = epm.out[0][k];
		++i;
		buff[i] = epm.out[1][k];
	}
}

/*----------------------------------------------------------------*/
void __stdcall effectProcessingInt32Mono(long *buff, long count, float freq)
{
	if (count < 0)return;
	EffectProcessingManager epm;

	for (int i = 0, k = 0; i < count; ++i, ++k) {// convert i32 -> f32
		epm.out[0][k] = epm.in[0][k] = 
			epm.out[1][k] = epm.in[1][k] =
				Sample_ConvI32toF(buff[i]);
	}

	effectProcessingBase(epm.in, epm.out, count, freq);

	for (int i = 0, k = 0; i < count; ++i,++k) {// convert f32 -> i32
		buff[i] = (Sample_ConvFtoI32(epm.out[1][k]) + Sample_ConvFtoI32(epm.out[0][k])) / 2;
	}
}

void __stdcall effectProcessingInt24Mono(char *buff, long count, float freq)
{
	if (count < 0)return;
	EffectProcessingManager epm;
	unsigned char *p;
	
	p = (unsigned char *)buff + 0;
	for (int k = 0; k < count; ++k) {
		epm.out[0][k] = epm.in[0][k] = 
			epm.out[1][k] = epm.in[1][k] = 
				Sample_Int24toFloat(p); p+=3;
	}

	effectProcessingBase(epm.in, epm.out, count, freq);

	p = (unsigned char *)buff + 0;
	for (int k = 0; k < count; ++k) {
		long l = Sample_FloatToInt24(epm.out[0][k]);
		long r = Sample_FloatToInt24(epm.out[1][k]);
		sint_32to24((l+r) / 2, p); p+=3;
	}
}

void __stdcall effectProcessingInt16Mono(short *buff, long count, float freq)
{
	if (count < 0)return;

	EffectProcessingManager epm;

	for (int k = 0; k < count; ++k) {// convert i32 -> f32
		epm.out[0][k] = epm.in[0][k] = 
			epm.out[1][k] = epm.in[1][k] =
				Sample_ConvI16toF(buff[k]);
	}

	effectProcessingBase(epm.in, epm.out, count, freq);

	for (int k = 0; k < count; ++k) {// convert f32 -> i32
		buff[k] = (Sample_ConvFtoI16(epm.out[1][k]) + Sample_ConvFtoI16(epm.out[0][k])) / 2;
	}
}

void __stdcall effectProcessingInt8Mono(char *buff, long count, float freq)
{
	if (count < 0)return;

	EffectProcessingManager epm;

	for (int i = 0, k = 0; i < count; ++i, ++k) {// convert i32 -> f32
		epm.out[0][k] = epm.in[0][k] = 
			epm.out[1][k] = epm.in[1][k] =
				Sample_ConvI8toF(buff[i]);
	}

	effectProcessingBase(epm.in, epm.out, count, freq);

	for (int i = 0, k = 0; i < count; ++i,++k) {// convert f32 -> i32
		buff[i] = (Sample_ConvFtoI8(epm.out[1][k]) + Sample_ConvFtoI8(epm.out[0][k])) / 2;
	}
}

void __stdcall effectInit()
{
/*
	VSTProcessingWaitLocker lk(g_processingLock);
	const int list_length = (int)PluginList.size();

	for (int i = 0; i < list_length; ++i) {
		AEffect *effect = PluginList[i].effect;
		if ( effect == NULL )continue;
		if ( ! PluginList[i].bSampleRateInitYes)continue;
		effectProcessing_ClearVstBuffer(effect);


	}
*/
	EffectProcessingManager epm;

	float **in = epm.in;
	float **out = epm.out;
	const int count = 1024;
	const int list_length = (int)PluginList.size();
	for ( int y = 0; y < 64; ++y) {
		VSTProcessingWaitLocker lk(g_processingLock);

		memset(WorkAudioBufferF[0], 0, sizeof(WorkAudioBufferF[0]));
		memset(WorkAudioBufferF[1], 0, sizeof(WorkAudioBufferF[1]));
		memset(WorkAudioBufferOutF[0], 0, sizeof(WorkAudioBufferOutF[0]));
		memset(WorkAudioBufferOutF[1], 0, sizeof(WorkAudioBufferOutF[1]));
		memset(channelDummyIn, 0, sizeof(channelDummyIn));
		memset(channelDummyOut, 0, sizeof(channelDummyOut));

		for (int i = 0; i < list_length; ++i) {
			AEffect *effect = PluginList[i].effect;
			if ( effect == NULL )continue;
			if ( ! PluginList[i].bSampleRateInitYes){ continue; }
			if ( ! PluginList[i].bProsessing) { continue; }
			mutex.lock();
			effect->processReplacing(effect, in, out, count);
			mutex.unlock();
		}
	}
}


//-------------------------------------------------------------------------------------------------------


VstIntPtr VSTCALLBACK HostCallback (AEffect* effect, VstInt32 opcode, VstInt32 index, VstIntPtr value, void* ptr, float opt)
{
	VstIntPtr result = 0;

	switch(opcode){
		case audioMasterIdle:
			effect->dispatcher(effect, effEditIdle, 0, 0, 0, 0);
			break;
		case audioMasterVersion :
			result = kVstVersion;
			break;
	}

	return result;
}


//-------------------------------------------------------------------------------------------------------
#include "resource.h"
HBITMAP hRes_ImgSave;
HBITMAP hRes_ImgLoad;

BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
	switch(fdwReason){
	case DLL_PROCESS_ATTACH:

		timeBeginPeriod(1);

		hDllInst = hinstDLL;
		GetModuleFileName(hDllInst, strDllPath, MAX_PATH);
		PathRemoveFileSpec(strDllPath);

		GetVstCurrentPath(strPresetPath, MAX_PATH);
		strcat(strPresetPath, "\\Preset");
		CreateDirectory(strPresetPath,NULL);

		hRes_ImgLoad = LoadBitmap(hDllInst, MAKEINTRESOURCE(IDB_VST_LOAD));
		hRes_ImgSave = LoadBitmap(hDllInst, MAKEINTRESOURCE(IDB_VST_SAVE));
		break;
	}
	return TRUE;
}

