#include "pch.h"
#include "common.h"
#include "tchar.h"
#include <OpenDlg.h>
#include <IniManager.h>
#define VSTC_OFFSET 46

INT_PTR CALLBACK EditorProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);
void savePluginPreset(PluginData *);
void loadPluginPreset(PluginData *);
void savePlugin_CustomPreset(const char *filename, AEffect *effect);
void loadPlugin_CustomPreset(const char *filename, AEffect *effect);

HWND hwndManager = NULL;

void getEffectName(PluginData *pd, char *str, size_t sz)
{
	const int tmp_buff_sz = 256;
	char effectName[tmp_buff_sz+2] = {0};
	pd->effect->dispatcher(pd->effect, effGetEffectName, 0, 0, effectName, 0);

	if (strcmp(effectName, "") == 0 || strcmp(effectName, " ") == 0){
		GetModuleFileName(pd->loader.module, effectName, tmp_buff_sz);
		strncpy(str, PathFindFileName( effectName), sz);
		int len = strlen(str);
		if (len > 4)str[len - 4] = 0;
		return;
	}
	strncpy(str, effectName, sz);
}

void InsertVSTItem(HWND hList, PluginData *pd, int idx)
{
	LV_ITEM lvi={0};
	lvi.mask = LVIF_TEXT | LVIF_STATE | LVIF_PARAM;

	char effectName[256] = {0};
	getEffectName(pd, effectName, 200);

	lvi.pszText = effectName;
	lvi.cchTextMax = strlen( effectName );
	lvi.iItem = ListView_GetItemCount(hList);
	lvi.iSubItem = 0;
	lvi.stateMask = LVIS_STATEIMAGEMASK;
	lvi.state = INDEXTOSTATEIMAGEMASK( pd->bProsessing ? 2: 1 );
	lvi.lParam = lvi.iItem;
	pd->index = lvi.iItem;

	ListView_InsertItem(hList, &lvi);

	ListView_SetCheckState(hList, idx, pd->bProsessing);
}

void SwapVSTItem(HWND hList, int dst, int src)
{
	TCHAR strDst[MAX_PATH+4] = _T("");
	TCHAR strSrc[MAX_PATH+4] = _T("");
	LVITEM ldst = {0}, lsrc = {0};
	lsrc.mask = ldst.mask = LVIF_TEXT | LVIF_STATE | LVIF_PARAM;
	lsrc.cchTextMax = ldst.cchTextMax = MAX_PATH;
	lsrc.pszText = strSrc; ldst.pszText = strDst;
	lsrc.iItem = src; ldst.iItem = dst;
	lsrc.iSubItem = ldst.iSubItem = 0;

	ListView_GetItem(hList, &ldst);
	ListView_GetItem(hList, &lsrc);

	int check1 = ListView_GetCheckState(hList, dst);
	int check2 = ListView_GetCheckState(hList, src);

	ldst.pszText = strSrc;
	lsrc.pszText = strDst;

	ListView_SetItem(hList, &ldst);
	ListView_SetItem(hList, &lsrc);

	ListView_SetCheckState(hList, dst, check2);
	ListView_SetCheckState(hList, src, check1);
}

void open_thr(PluginData *pd)
{
//		pd->hwnd = CreateDialogParamA(hDllInst, MAKEINTRESOURCE(IDD_VSTCTRL), 0, (DLGPROC)EditorProc, (LPARAM)pd);
		DialogBoxParamA(hDllInst, MAKEINTRESOURCE(IDD_VSTCTRL), 0, (DLGPROC)EditorProc, (LPARAM)pd);
}

bool openEffectEditor(PluginData *pd, HWND hwnd, bool createOnly = false)
{
	if ((pd->effect->flags & effFlagsHasEditor) == 0){
//		return false;
	}

	if (pd->hwnd == NULL) {
		pd->hwnd = CreateDialogParamA(hDllInst, MAKEINTRESOURCE(IDD_VSTCTRL), 0, (DLGPROC)EditorProc, (LPARAM)pd);
		if (!createOnly)
			ShowWindow(pd->hwnd, SW_SHOW);
		return true;
	}

	if (createOnly)return true;

	if (pd->bShow) {
		ShowWindow(pd->hwnd, SW_SHOW);
		SetFocus( pd->hwnd );
	}else{
		ShowWindow(pd->hwnd, SW_SHOW);
		pd->bShow = true;
	}
	return true;
}

void updatePluginDataPointer_unsafe(PluginData *ptr)
{
	RemoveProp(ptr->hwnd, "__VST_EDIT__");
	SetProp(ptr->hwnd, "__VST_EDIT__", ptr);
}

void updatePluginDataPointer(PluginData *ptr)
{
	boost::mutex::scoped_lock lk(mutex);

	RemoveProp(ptr->hwnd, "__VST_EDIT__");
	SetProp(ptr->hwnd, "__VST_EDIT__", ptr);
}

PluginData *getPluginData(HWND hwnd)
{
	return reinterpret_cast<PluginData *>(GetProp(hwnd, "__VST_EDIT__"));
}

extern HBITMAP hRes_ImgSave;
extern HBITMAP hRes_ImgLoad;
extern char *script_spaceskip(char *ptr);
#define VST_IDEL_TIMER 535

extern void loadChunk(PluginData *pld, const char *fn);
extern void saveChunk(PluginData *pld, const char *fn);

void CustomPreset_SaveThread(const HWND hwnd, PluginData *pld)
{
	AEffect *effect = pld->effect;
	TCHAR fn[MAX_PATH + 4]=_T("custom.pri");

	TCHAR currentDir[MAX_PATH+4]=_T("");
	GetCurrentDirectory(MAX_PATH, currentDir);

	mutex.lock();
	OPENFILENAME ofn;
	memset(&ofn,0,sizeof(OPENFILENAME)); 
	ofn.lStructSize = sizeof(OPENFILENAME); 
	ofn.hwndOwner = NULL;

	if (effect->flags & effFlagsProgramChunks) {
		ofn.lpstrFilter = _T(
			"VST Preset File (*.pri)\0*.pri\0" \
			"VST Plugin Bank (*.fxb)\0*.fxb\0"\
			"VST Plugin Program (*.fxp)\0*.fxp\0" \
			"\0"
		);
	}else{
		ofn.lpstrFilter = _T("VST Preset File (*.pri)\0*.pri\0\0");
	}
	ofn.lpstrTitle = _T("Save Custom Preset");
	ofn.lpstrDefExt = _T(".pri");
	ofn.nFilterIndex = 1; 
	ofn.lpstrFile = fn; 
	ofn.nMaxFile = MAX_PATH; 
	ofn.Flags = OFN_CREATEPROMPT|OFN_OVERWRITEPROMPT|OFN_NOCHANGEDIR; 
	mutex.unlock();

	if(GetSaveFileName(&ofn)) { 
		mutex.lock();
		if (stricmp(PathFindExtension(fn), ".pri") == 0) {
			savePlugin_CustomPreset(fn, effect);
		}else{
			saveChunk(pld, fn);
		}
		mutex.unlock();
	}

	SetCurrentDirectory(currentDir);
}

void CustomPreset_Save(HWND hwnd, PluginData *pld)
{
	boost::thread(boost::bind(CustomPreset_SaveThread, hwnd, pld));
}

void CustomPreset_LoadThread(const HWND hwnd, PluginData *pld)
{
	AEffect *effect = pld->effect;

	TCHAR fn[MAX_PATH + 4]=_T("custom.pri");

	TCHAR currentDir[MAX_PATH+4]=_T("");
	GetCurrentDirectory(MAX_PATH, currentDir);

	mutex.lock();
	OPENFILENAME ofn;
	memset(&ofn,0,sizeof(OPENFILENAME)); 
	ofn.lStructSize = sizeof(OPENFILENAME); 
	ofn.hwndOwner = NULL; 
	if (effect->flags & effFlagsProgramChunks) {
		ofn.lpstrFilter = _T(
			"VST Preset File (*.pri)\0*.pri\0" \
			"VST Plugin Bank (*.fxb)\0*.fxb\0"\
			"VST Plugin Program (*.fxp)\0*.fxp\0" \
			"\0"
		);
	}else{
		ofn.lpstrFilter = _T("VST Preset File (*.pri)\0*.pri\0\0");
	} 
	ofn.lpstrTitle = _T("Load Custom Preset (*.pri)");
	ofn.lpstrDefExt = _T(".pri");
	ofn.nFilterIndex = 1; 
	ofn.lpstrFile = fn; 
	ofn.nMaxFile = MAX_PATH; 
	ofn.Flags = OFN_FILEMUSTEXIST|OFN_NOCHANGEDIR; 
	mutex.unlock();

	if ( GetOpenFileName(&ofn) ) {
		mutex.lock();
		if (stricmp(PathFindExtension(fn), ".pri") == 0) {
			loadPlugin_CustomPreset(fn, effect);
		} else loadChunk(pld, fn);
		mutex.unlock();
	}
	SetCurrentDirectory(currentDir);
}

void CustomPreset_Load(HWND hwnd, PluginData *pld)
{
	boost::thread(boost::bind(CustomPreset_LoadThread, hwnd, pld));
}

extern HWND CreateCustomEditorDialog(HWND hwnd, PluginData *pd);

void UpdatePresetBox(HWND hwnd, UINT id, int &w, int height)
{
	RECT rc;
	GetClientRect(GetDlgItem(hwnd, id), &rc);
//	AdjustWindowRectEx(&rc, GetWindowLong (GetDlgItem(hwnd, id), GWL_STYLE), FALSE, GetWindowLong (GetDlgItem(hwnd, id), GWL_EXSTYLE));
	SetWindowPos(GetDlgItem(hwnd, id), HWND_TOP, rc.left+w, height-20, rc.right-rc.left, 20, SWP_NOSIZE | SWP_NOZORDER );
	w+=rc.right;
}

void EditorProc_wmInitDialog(__in const HWND hwnd, __in const WPARAM wParam, __in const LPARAM lParam)
{
	SetProp(hwnd, "__VST_EDIT__", (HANDLE)lParam);

	AEffect *effect = getPluginData(hwnd)->effect;

	char effectName[256] = {0};
	getEffectName(getPluginData(hwnd), effectName, 200);

	char winName[256];
	sprintf(winName, "timvst: %s", effectName);
	SetWindowText (hwnd, winName);

	SendDlgItemMessage(hwnd, IDC_EC_BTNPRESET_LOAD, BM_SETIMAGE, IMAGE_BITMAP, (LPARAM)hRes_ImgLoad);
	SendDlgItemMessage(hwnd, IDC_EC_BTNPRESET_SAVE, BM_SETIMAGE, IMAGE_BITMAP, (LPARAM)hRes_ImgSave);

	if (!effect)return;

	ERect* eRect = 0;
	ERect eCutomEdit;

	for (int i = 0; i < effect->numPrograms; ++i){
		char progName[256] = {0};
		if (!effect->dispatcher(effect, effGetProgramNameIndexed, i, 0, progName, 0)){
			effect->dispatcher(effect, effSetProgram, 0, i, 0, 0); // Note: old program not restored here!
			effect->dispatcher(effect, effGetProgramName, 0, 0, progName, 0);
		}
		ComboBox_AddString(GetDlgItem(hwnd, IDC_EC_CBPRESET), progName);
	}
	ComboBox_SetCurSel(GetDlgItem(hwnd, IDC_EC_CBPRESET), 0);
	if (!getPluginData(hwnd)->bChannelVstMode){
		loadPluginPreset( getPluginData(hwnd) );
	}

	if ((effect->flags & effFlagsHasEditor) == 0){
		HWND hchild = CreateCustomEditorDialog(hwnd, getPluginData(hwnd));
		RECT rc;
		GetWindowRect(hchild, &rc);
		eCutomEdit.top = (short)rc.top;
		eCutomEdit.bottom = (short)rc.bottom;
		eCutomEdit.left = (short)rc.left;
		eCutomEdit.right = (short)rc.right;
		eRect = &eCutomEdit;
	}else{
		effect->dispatcher (effect, effEditOpen, 0, 0, hwnd, 0);
		effect->dispatcher (effect, effEditGetRect, 0, 0, &eRect, 0);
	}

	int width, height;
	if (eRect){
		width = eRect->right - eRect->left;
		height = eRect->bottom - eRect->top;
		if (width < 100)
			width = 100;
		if (height < 100)
			height = 100;

		RECT wRect;
		SetRect(&wRect, 0, 0, width, height);
		AdjustWindowRectEx (&wRect, GetWindowLong (hwnd, GWL_STYLE), FALSE, GetWindowLong (hwnd, GWL_EXSTYLE));
		width = wRect.right - wRect.left;
		height = wRect.bottom - wRect.top;

//		SetWindowPos (hwnd, HWND_TOP, 0, 20, width, height, SWP_NOMOVE);
		SetWindowPos (hwnd, HWND_TOP, 0, 0, width, height + 24, SWP_NOMOVE);

		int pretset_w = 0;
		UpdatePresetBox(hwnd,IDC_EC_CBPRESET,pretset_w,height);
		UpdatePresetBox(hwnd,IDC_EC_BTNPRESET_SAVE,pretset_w,height);
		UpdatePresetBox(hwnd,IDC_EC_BTNPRESET_LOAD,pretset_w,height);
	}

	if (effect) {
		SetTimer(hwnd, VST_IDEL_TIMER, 20, 0);
	}
}

void EditorProc_wmCommand(__in const HWND hwnd, __in const WPARAM wParam, __in const LPARAM lParam)
{
	switch( LOWORD(wParam) ) {
	case IDC_EC_CBPRESET:{
		if( HIWORD(wParam) == CBN_SELCHANGE ) {
			HWND hcb = GetDlgItem(hwnd, IDC_EC_CBPRESET);
			int idx = ComboBox_GetCurSel(hcb);
			if(idx == -1)break;
			AEffect *effect = getPluginData(hwnd)->effect;
			effect->dispatcher (effect, effSetProgram, 0, idx, 0, 0);
			SetFocus(GetDlgItem(hwnd, IDC_EC_PASTEAREA));
			return;
		}
		return;
	}
	case IDC_EC_BTNPRESET_LOAD:
	{
		CustomPreset_Load(hwnd, getPluginData(hwnd));
		return;
	}
	case IDC_EC_BTNPRESET_SAVE:
	{
		CustomPreset_Save(hwnd, getPluginData(hwnd));
		return;
	}
	default:
		return;
	}
}

void EditorProc_wmTimer(__in const HWND hwnd, __in const WPARAM wParam, __in const LPARAM lParam)
{
	switch (wParam){
	case VST_IDEL_TIMER: {
		PluginData *pd = getPluginData(hwnd);
		if (pd == NULL)return;
		if (pd->effect) {
			pd->effect->dispatcher(pd->effect, effEditIdle, 0, 0, 0, 0);
		}
		break;
	}
	default:
		break;
	}
}

void EditorProc_wmClose(__in const HWND hwnd, __in const WPARAM wParam, __in const LPARAM lParam)
{
	PluginData *pd = getPluginData(hwnd);
	AEffect *effect = pd->effect;

	if (!pd->bChannelVstMode){
		savePluginPreset(pd);
	}
	pd->bShow = false;

	ShowWindow(hwnd, SW_HIDE);

	if (pd->bSystemEnd && pd->hwnd != NULL) {
		KillTimer (hwnd, VST_IDEL_TIMER);
		if (effect)
			effect->dispatcher (effect, effEditClose, 0, 0, 0, 0);
		RemoveProp(hwnd, "__VST_EDIT__");

		DestroyWindow(hwnd);
		pd->hwnd = NULL;
	}
}

INT_PTR CALLBACK EditorProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch(msg){
		case WM_INITDIALOG:
			EditorProc_wmInitDialog(hwnd,wParam,lParam);
			return TRUE;
		case WM_COMMAND:
			EditorProc_wmCommand(hwnd,wParam,lParam);
			return FALSE;
		case WM_TIMER:
			EditorProc_wmTimer(hwnd,wParam,lParam);
			return FALSE;
		case WM_CLOSE:
			EditorProc_wmClose(hwnd,wParam,lParam);
			return FALSE;
		case WM_DESTROY:
			break;
	}

	return 0;
}

// ---------------------------------------------------------------
// 
// plugin manager
//
// ---------------------------------------------------------------

int GetListIndex(HWND hList)
{
	int n = -1;
	while((n=ListView_GetNextItem(hList, n, LVNI_SELECTED)) != -1)
		return n;
	return -1;
}

void WaitMsgLoop(WORD Count, BYTE SleepCount)
{
	MSG msg = {0};
	if (!Count) {
		if (PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE)) {
			if (!GetMessage(&msg, NULL, 0, 0))
				return;
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
		return;
	}

	for (WORD i = 0; i < Count; ++i) {
		if (PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE)) {
			if (!GetMessage(&msg, NULL, 0, 0))
				return;
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}else{
			Sleep(SleepCount);
		}
	}
}

#ifndef ListView_SetCheckState
   #define ListView_SetCheckState(hwndLV, i, fCheck) \
      ListView_SetItemState(hwndLV, i, \
      INDEXTOSTATEIMAGEMASK((fCheck)+1), LVIS_STATEIMAGEMASK)
#endif

RECT GetWindowPosition(HWND hwnd)
{
	WINDOWPLACEMENT wp;
	GetWindowPlacement(hwnd, &wp);
	return wp.rcNormalPosition;
}



int g_VstLoaderIndex = 0;
HMENU g_hVstItemMenu;
void EmptyFunction(){ MessageBox(NULL, "vst open err", "err", MB_OK | MB_ICONERROR); }
TCHAR g_VstFolderPath[MAX_PATH + 4];

void CreateVSTItemMenu()
{
	IniMemFile ini;
	ini.LoadFromFile("soundfont.ini");
	ini.ReadCStr("VST", "Path", g_VstFolderPath, MAX_PATH, "C:\\Program Files\\VST");

	TCHAR *p = g_VstFolderPath;
	WIN32_FIND_DATA fd;
	TCHAR searchFolder[MAX_PATH+4];
	strcpy(searchFolder, g_VstFolderPath);
	strcat(searchFolder, "\\*.dll");
	HANDLE hfd = FindFirstFile(searchFolder, &fd);
	g_hVstItemMenu = CreatePopupMenu();
	if (hfd == INVALID_HANDLE_VALUE) {
	}else{
		unsigned int i = 0;
		do {
			InsertMenu(g_hVstItemMenu, 0, MF_BYPOSITION|MF_STRING, (UINT)(VST_CHANNEL_START_ID + i), PathFindFileName(fd.cFileName));
			i++;
		}while(FindNextFile(hfd,&fd));
		FindClose(hfd);
	}
}
void CreateChannelControl(HWND hwnd)
{
	HWND hBaseStatic = GetDlgItem(hwnd, IDC_VW_LB00);
	DWORD dwStaticStyle = GetWindowLong(hBaseStatic, GWL_STYLE);
	RECT stStaticRect = GetWindowPosition(hBaseStatic);
	HFONT hFont = (HFONT)SendMessage(hwnd, WM_GETFONT, 0, 0);

	HWND hBaseChkBox = GetDlgItem(hwnd, IDC_VW_CHK00);
	DWORD dwChkBoxStyle = GetWindowLong(hBaseChkBox, GWL_STYLE);
	RECT stChkBoxRect = GetWindowPosition(hBaseChkBox);

	HWND hBaseButton = GetDlgItem(hwnd, IDC_VW_BTN00);
	DWORD dwButtonStyle = GetWindowLong(hBaseButton, GWL_STYLE);
	RECT stButtonRect = GetWindowPosition(hBaseButton);

	HWND hBaseButton2 = GetDlgItem(hwnd, IDC_VW_BTNLOAD_00);
	DWORD dwButtonStyle2 = GetWindowLong(hBaseButton2, GWL_STYLE);
	RECT stButtonRect2 = GetWindowPosition(hBaseButton2);

	WINDOWPLACEMENT wp;
	GetWindowPlacement(hwnd, &wp);
	wp.rcNormalPosition.bottom += 34 * (g_MaxChannelVstCount - 1) + 5;
	SetWindowPlacement(hwnd, &wp);

	for (int k = 0, set_i = 0; k < g_MaxChannelVstCount; ++k) {
		ChannelVstItem[k].ID_VSTInsertBtn = IDC_VW_BTN00 + ( k * VSTC_OFFSET );
		ChannelVstItem[k].ID_VSTInsertBtn_VstLoad = IDC_VW_BTNLOAD_00 + ( k * VSTC_OFFSET );

		if (k == 0) {
			set_i = 1;
		}else{
			set_i = 0;

			HWND hc = CreateWindow("BUTTON","No VST", dwButtonStyle,
				stButtonRect.left, k * 34 + stButtonRect.top, stButtonRect.right- stButtonRect.left, 30, hwnd, (HMENU)ChannelVstItem[k].ID_VSTInsertBtn, hDllInst, NULL);
			SendMessage(hc, WM_SETFONT, (WPARAM)hFont, 1);
			hc = CreateWindow("BUTTON","add", dwButtonStyle2,
				stButtonRect2.left, k * 34 + stButtonRect2.top, stButtonRect2.right- stButtonRect2.left, 30, hwnd, (HMENU)ChannelVstItem[k].ID_VSTInsertBtn_VstLoad, hDllInst, NULL);
			SendMessage(hc, WM_SETFONT, (WPARAM)hFont, 1);
		}

		if (strcmp(ChannelVstItem[k].name,"")){
			SetWindowText(GetDlgItem(hwnd, ChannelVstItem[k].ID_VSTInsertBtn), ChannelVstItem[k].name);
		}

		const unsigned long id_inc_chk = IDC_VW_CHK00 + (k * VSTC_OFFSET);

		for (unsigned long i = set_i; i < 32; ++i) {
			char buf[32];
			sprintf(buf, "%d", i+1);

			HWND hChild;
			hChild = CreateWindowA("STATIC", buf, dwStaticStyle, 
				stStaticRect.left + i*14, k * 34 + stStaticRect.top, 14, 14, hwnd, (HMENU)-1, hDllInst, 0);
			SendMessage(hChild, WM_SETFONT, (WPARAM)hFont, 1);

			hChild = CreateWindowA("BUTTON", buf, dwChkBoxStyle, 
				stChkBoxRect.left + i*14, k * 34 + stChkBoxRect.top, 14, 14, hwnd, (HMENU)(id_inc_chk + i), hDllInst, 0);
			SendMessage(hChild, WM_SETFONT, (WPARAM)hFont, 1);
		}
		for (unsigned long i = 0; i < 32; ++i) {
			CheckDlgButton(hwnd, id_inc_chk + i, (ChannelVstItem[k].chMask & (1<<i)) ? BST_CHECKED : BST_UNCHECKED);
		}
	}
	
	CreateVSTItemMenu();
}

void PluginList_DeleteEndBlock_Unsafe()
{
	PluginList.erase( PluginList.end()-1 );
}

extern bool VST_Loader( PluginData *pld, char *path, void (* PluginListBackDeleter)());

struct PathBuffer
{
	TCHAR path[MAX_PATH + 4];
	PathBuffer()
	{
		path[0] = 0;
	}

	PathBuffer(LPCTSTR p)
	{
		_tcscpy(path,p);
	}

	LPCTSTR operator=(LPCTSTR)
	{
		return path;
	}

	LPCTSTR c_str()
	{
		return path;
	}
};

void ChannelVSTLoadThread(HWND hwnd, int vst_idx, PathBuffer svstName, PathBuffer VstPath)
{
	TCHAR openVstPath[MAX_PATH+4];
	strcpy(openVstPath,VstPath.c_str());

	TCHAR vstName[MAX_PATH+4];
	strcpy(vstName,svstName.c_str());
}

static bool bRunChannelProcessing = false;

void ChannelManager_Command(HWND hwnd, const unsigned short CtrlId, const LPARAM lp)
{
	if (CtrlId < 6000)return;

	if (VST_CHANNEL_START_ID <= CtrlId && VST_CHANNEL_END_ID >= CtrlId) {
		
		TCHAR openVstPath[MAX_PATH + 4];
		TCHAR vstName[MAX_PATH + 4];
		GetMenuString(g_hVstItemMenu, CtrlId, vstName, MAX_PATH, MF_BYCOMMAND);
		strcpy(openVstPath,g_VstFolderPath);
		strcat(openVstPath,"\\");
		strcat(openVstPath,vstName);

		if (g_VstLoaderIndex == -1) {
			mutex.lock();
			PluginList.push_back(PluginData());
			PluginData &p = PluginList.back();

			if (!VST_Loader(&p, openVstPath, PluginList_DeleteEndBlock_Unsafe)) {
				mutex.unlock();
				return;

			}
			HWND hList = GetDlgItem(hwnd,IDC_LIST1);
			p.bProsessing = false;
			InsertVSTItem(hList, &p, ListView_GetItemCount(hList));
			mutex.unlock();

			openEffectEditor(&p, NULL, true);

			mutex.lock();
			for (size_t i = 0; i < PluginList.size(); ++i) {
				updatePluginDataPointer_unsafe(&PluginList[i]);
			}
			mutex.unlock();

			return;
		}

		if (bRunChannelProcessing) {
			MessageBox(hwnd, "łB\r\nĎsĂ", "G[", MB_OK|MB_ICONERROR);
			return;
		}

		mutex.lock();

		const long vst_idx = g_VstLoaderIndex;
		ChannelVstItem[vst_idx].chMask = 0;
		for (int i = 0; i < 32; ++i) {
			CheckDlgButton(hwnd, IDC_VW_CHK00 + (vst_idx * VSTC_OFFSET) + i, BST_UNCHECKED);
		}

		ChannelVstItem[ vst_idx ].pd.bSystemEnd = true;
		if (ChannelVstItem[ vst_idx ].pd.hwnd != NULL) {
			SendMessage(ChannelVstItem[ vst_idx ].pd.hwnd, WM_CLOSE, 0, 0);
		}

		ChannelVstItem[ vst_idx ].pd.clear();

		ChannelVstItem[ vst_idx ].pd.bChannelVstMode = true;
		VST_Loader(&ChannelVstItem[ vst_idx ].pd, openVstPath, EmptyFunction);
		if (ChannelVstItem[ vst_idx ].pd.effect != NULL) {
			SetWindowText(GetDlgItem(hwnd, IDC_VW_BTN00 + (vst_idx * VSTC_OFFSET)), vstName); 
			strncpy(ChannelVstItem[ vst_idx ].name, vstName, (sizeof(ChannelVstItem[ vst_idx ].name)-1)/sizeof(TCHAR));
			openEffectEditor(&ChannelVstItem[ vst_idx ].pd, NULL, true);
			ChannelVstItem[ vst_idx ].pd.bProsessing = true;
		}
		mutex.unlock();

		return;
	}
	unsigned int k = IDC_VW_CHK00;
	for (int i = 0; i < g_MaxChannelVstCount; ++i) {
		if (CtrlId == IDC_VW_BTNLOAD_00 + (i * VSTC_OFFSET)) {
			g_VstLoaderIndex = i;
			POINT pt;
			GetCursorPos(&pt);
			TrackPopupMenu( g_hVstItemMenu, TPM_LEFTALIGN, pt.x, pt.y, 0, hwnd, NULL );
			return;
		}
		if (CtrlId == IDC_VW_BTN00 + (i * VSTC_OFFSET)) {
			if(ChannelVstItem[ i ].pd.effect)
			openEffectEditor(&ChannelVstItem[ i ].pd, NULL);
			return;
		}
		if (CtrlId >= k && CtrlId <k + (VSTC_OFFSET - 3)) {
			unsigned long chk_index = CtrlId - k;
			int dlgItemChecked = IsDlgButtonChecked(hwnd, CtrlId);

			mutex.lock();
			if ( dlgItemChecked == BST_CHECKED ) {
				ChannelVstItem[i].chMask |= (1<<chk_index);
			}else{
				ChannelVstItem[i].chMask &= ~(1<<chk_index);
			}
			mutex.unlock();

			return;
		}

		k += VSTC_OFFSET;
	}
}

static float channelWorkBuffIn[2][CHANNEL_AUDIO_BUFF] = {0};
static float channelWorkBuffOut[2][CHANNEL_AUDIO_BUFF] = {0};

static long VSTEffectBuff[CHANNEL_AUDIO_BUFF] = {0};
static long VSTEffectTmpBuff[CHANNEL_AUDIO_BUFF] = {0};
extern long Sample_ConvFtoI32(float v);
extern float Sample_ConvI32toF(long v);
extern void effectProcessing_ClearVstBuffer(AEffect *);


float channelDummyIn[MAX_AUDIO_BUFFER] = {0};
float channelDummyOut[MAX_AUDIO_BUFFER] = {0};

void effectChannelProcessingBase(const unsigned long channel_mask, float **in, float **out, long count, float freq)
{
	const unsigned long countbyte = ( sizeof(float) * (count) );

	for (int i = 0; i < g_MaxChannelVstCount; ++i) {

		AEffect *effect = ChannelVstItem[ i ].pd.effect;
		if (effect == NULL)continue;
		if (ChannelVstItem[ i ].pd.bSystemEnd)
			continue;
		if (!(ChannelVstItem[i].chMask & channel_mask))
			continue;

		if (ChannelVstItem[ i ].pd.bClearVstBuffer) {// buffer init!
			ChannelVstItem[ i ].pd.bClearVstBuffer = false;
			effectProcessing_ClearVstBuffer(effect);
		}

		if ( ! ChannelVstItem[ i ].pd.bSampleRateInitYes)
			continue;

		boost::mutex::scoped_lock lk(mutex);

		if (effect->numInputs == 1) { // mono vst !
			for (int ib = 0; ib < count; ++ib) {
				in[1][ib] = in[0][ib] = (in[0][ib] + in[1][ib]) / 2.0f;
			}
		}

		effect->processReplacing(effect, in, out, count);

		if (effect->numOutputs == 1) { // mono vst !
			memcpy(in[0], out[0], countbyte);
			memcpy(in[1], out[0], countbyte);
		}else{
			memcpy(in[0], out[0], countbyte);
			memcpy(in[1], out[1], countbyte);
		}
	}
}

void __stdcall effectChannelProcessingInt32(const unsigned long channel, long *buff, long count, float freq)
{
	if (count < 0)return;

	bRunChannelProcessing = true;

	if (channel == 0) {
		static float svFreq = 0;
		if (svFreq != freq) {
			svFreq = freq;
			for (int i = 0; i < g_MaxChannelVstCount; ++i) {
				if (ChannelVstItem[ i ].pd.bSampleRateInitYes)
					ChannelVstItem[ i ].pd.effect->dispatcher(ChannelVstItem[ i ].pd.effect, effSetSampleRate, 0, 0, 0, svFreq);
			}
		}
		
		for (int i = 0; i < g_MaxChannelVstCount; ++i) {

			if (ChannelVstItem[ i ].pd.bSystemEnd)
				continue;

			boost::mutex::scoped_lock lk(mutex);

			AEffect *effect = ChannelVstItem[ i ].pd.effect;
			if ( ! ChannelVstItem[ i ].pd.effect)continue;
			if ( ! ChannelVstItem[ i ].pd.bSampleRateInitYes) { // init !
				effect->dispatcher(effect, effSetSampleRate, 0, 0, 0, svFreq);
				effect->dispatcher(effect, effSetBlockSize, 0, CHANNEL_AUDIO_BUFF, 0, 0);
				effect->dispatcher(effect, effMainsChanged, 0, 1, 0, 0);
				
				ChannelVstItem[ i ].pd.bSampleRateInitYes = true;
			}
		}
	}

	const unsigned long mask_channel = 1UL<<channel;
	{
		int chk  = 0;
		for (; chk < g_MaxChannelVstCount; ++chk) {
			if (ChannelVstItem[chk].chMask & mask_channel) {
				break;
			}
		}
		if (chk == g_MaxChannelVstCount) {
			bRunChannelProcessing = false;
			return;
		}
	}

	float *in[MAX_INOUT_CHANNEL] = {channelWorkBuffIn[0],channelWorkBuffIn[1]};
	float *out[MAX_INOUT_CHANNEL] = {channelWorkBuffOut[0],channelWorkBuffOut[1]};
	for (int i = 2; i < MAX_INOUT_CHANNEL; ++i) {
		in[i] = channelDummyIn;
		out[i] = channelDummyOut;
	}


	for (int i = 0, k = 0; i < count; ++i, ++k) {// convert i32 -> f32
		out[0][k] = in[0][k] = Sample_ConvI32toF(buff[i]);
		++i;
		out[1][k] = in[1][k] = Sample_ConvI32toF(buff[i]);
	}

	effectChannelProcessingBase(mask_channel, in, out, count / 2, freq);

	for (int i = 0, k = 0; i < count; ++i,++k) {// convert f32 -> i32
		buff[i] = Sample_ConvFtoI32(out[0][k]);
		++i;
		buff[i] = Sample_ConvFtoI32(out[1][k]);
	}

	bRunChannelProcessing = false;
}



void effectChannelProcessingBaseB(long i, float **in, float **out, long count, float freq)
{
	const unsigned long countbyte = ( sizeof(float) * (count) );

	{
		AEffect *effect = ChannelVstItem[ i ].pd.effect;
		if (effect == NULL)return;
		if (ChannelVstItem[ i ].pd.bSystemEnd)
			return;

		if ( ! ChannelVstItem[ i ].pd.bSampleRateInitYes)
			return;

		if (effect->numInputs == 1) { // mono vst !
			for (int ib = 0; ib < count; ++ib) {
				in[1][ib] = in[0][ib] = (in[0][ib] + in[1][ib]) / 2.0f;
			}
		}

		effect->processReplacing(effect, in, out, count);

		if (effect->numOutputs == 1) { // mono vst !
			memcpy(in[0], out[0], countbyte);
			memcpy(in[1], out[0], countbyte);
		}else{
			memcpy(in[0], out[0], countbyte);
			memcpy(in[1], out[1], countbyte);
		}
	}
}

void effectChannelProcessingInt32B(int effId, long count, float freq)
{
	bRunChannelProcessing = true;

	float *in[MAX_INOUT_CHANNEL] = {channelWorkBuffIn[0],channelWorkBuffIn[1]};
	float *out[MAX_INOUT_CHANNEL] = {channelWorkBuffOut[0],channelWorkBuffOut[1]};
	for (int i = 2; i < MAX_INOUT_CHANNEL; ++i) {
		in[i] = channelDummyIn;
		out[i] = channelDummyOut;
	}

	for (int i = 0, k = 0; i < count; ++i, ++k) {// convert i32 -> f32
		out[0][k] = in[0][k] = Sample_ConvI32toF(VSTEffectBuff[i]);
		++i;
		out[1][k] = in[1][k] = Sample_ConvI32toF(VSTEffectBuff[i]);
	}

	effectChannelProcessingBaseB(effId, in, out, count / 2, freq);

	for (int i = 0, k = 0; i < count; ++i,++k) {// convert f32 -> i32
		VSTEffectBuff[i] = Sample_ConvFtoI32(out[0][k]);
		++i;
		VSTEffectBuff[i] = Sample_ConvFtoI32(out[1][k]);
	}

	bRunChannelProcessing = false;
}

static long InOutResultBuffer[ CHANNEL_AUDIO_BUFF ] ={0};
long * WINAPI channelEffect_GetInOutResultBuffer()
{
	return InOutResultBuffer;
}
#define WIN_64

void SetMixInOut(long *buf, long count)
{
#ifdef _WIN64
	register long *dbuf = InOutResultBuffer;

	for(int i = count - 1; i >= 0; i--)
	{
		dbuf[i] += buf[i];
	}
#else
	long *dbuf = InOutResultBuffer;
	_asm {
		mov		ecx, [count]
		mov		esi, [buf]
		test	ecx, ecx
		jz		short L2
		mov		edi, [dbuf]
L1:		mov		eax, [esi]
		mov		ebx, [edi]
		add		esi, 4
		add		ebx, eax
		mov		[edi], ebx
		add		edi, 4
		dec		ecx
		jnz		L1
L2:
	}
#endif
	
}

void WINAPI channelEffect_InOut(long **buff, long count, float freq, long *direct_buff_ptr)
{
	if (count <= 0)return;
	if (buff == NULL)return;

	memset(InOutResultBuffer,0,sizeof(long)*count);

	static float svFreq = 1000;
	if (svFreq != freq) {
		svFreq = freq;

		boost::mutex::scoped_lock lk(mutex);

		for (int i = 0; i < g_MaxChannelVstCount; ++i) {
			if (ChannelVstItem[ i ].pd.bSampleRateInitYes)
				ChannelVstItem[ i ].pd.effect->dispatcher(ChannelVstItem[ i ].pd.effect, effSetSampleRate, 0, 0, 0, freq);
		}
	}

	for (int i = 0; i < g_MaxChannelVstCount; ++i) {

		if (ChannelVstItem[ i ].pd.bSystemEnd)
			continue;

		boost::mutex::scoped_lock lk(mutex);

		AEffect *effect = ChannelVstItem[ i ].pd.effect;
		if ( ! ChannelVstItem[ i ].pd.effect)continue;
		if ( ! ChannelVstItem[ i ].pd.bSampleRateInitYes) { // init !
			effect->dispatcher(effect, effSetSampleRate, 0, 0, 0, svFreq);
			effect->dispatcher(effect, effSetBlockSize, 0, CHANNEL_AUDIO_BUFF, 0, 0);
			effect->dispatcher(effect, effMainsChanged, 0, 1, 0, 0);
			
			ChannelVstItem[i].pd.bSampleRateInitYes = true;
		}
	}

	unsigned long vstUsedAllChannel = 0;
	for (int ef = 0; ef < g_MaxChannelVstCount; ++ef) {

		boost::mutex::scoped_lock lk(mutex);

		// in
		vstUsedAllChannel |= ChannelVstItem[ef].chMask;
		for (unsigned long n = 0; n < 32; n++) {

			const unsigned long mask_channel = 1UL<<n;
			if (ChannelVstItem[ef].chMask & mask_channel) {
				for (int i = 0; i < count; ++i)
					VSTEffectBuff[i]+=buff[n][i];
			}
		}
		// prosessing
		if (ChannelVstItem[ef].chMask != 0) {
			effectChannelProcessingInt32B(ef, count, freq);
		}

		// output
		for (int i = 0; i < count; ++i) {
			VSTEffectTmpBuff[i] += VSTEffectBuff[i];
		}

		memset(VSTEffectBuff, 0, sizeof(long)*count);
	}

	boost::mutex::scoped_lock lk(mutex);

	for (int i = 0; i < count; ++i) {
		InOutResultBuffer[i] += VSTEffectTmpBuff[i];
	}

	// no-vst effect mixing!
	for (unsigned long n = 0; n < 32; n++) {
		const unsigned long mask_channel = 1UL<<n;
		if (!(vstUsedAllChannel & mask_channel)) {
			if (direct_buff_ptr != buff[n] && buff[n] && !IsBadCodePtr((FARPROC)buff[n])){
				SetMixInOut(buff[n],count);
			}
		}
	}

	memset(VSTEffectTmpBuff, 0, sizeof(long)*count);
}


void ShowMsg2(const HWND hWnd, LPCTSTR Text, ...)
{
	TCHAR szBuffer[2048] = _T("");

	va_list marker;
	va_start(marker, Text);
	_vstprintf(szBuffer, Text, marker);
	va_end(marker);

	MessageBox(hWnd, szBuffer, _T("Show Message"), MB_OK);
}

long GetOSVerInt()
{
	OSVERSIONINFOEX OSInfo = {0};
	OSInfo.dwOSVersionInfoSize = sizeof(OSVERSIONINFOEX);
	int ret = GetVersionEx((OSVERSIONINFO *)&OSInfo);

	if (ret == 0) {
		OSVERSIONINFO OldOSInfo = {0};
		OldOSInfo.dwOSVersionInfoSize = sizeof OSVERSIONINFO;
		int platid = OldOSInfo.dwPlatformId;
		if(platid == VER_PLATFORM_WIN32s) 
			return 000;
		else if(platid == VER_PLATFORM_WIN32_WINDOWS){
			if(OSInfo.dwMinorVersion==0)
				return 1000;
			if(OSInfo.dwMinorVersion==10)
				return 2000;
			return 3000;
		}
		return (unsigned long)-1;
	}

	if (OSInfo.dwPlatformId == VER_PLATFORM_WIN32s){
		return 0000;
	}

	if (OSInfo.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS) {
		if(OSInfo.dwMinorVersion == 0)
			return 1000;
		if(OSInfo.dwMinorVersion == 10)
			return 2000;
		return 3000;
	}

	return (OSInfo.dwMajorVersion * 1000) + (OSInfo.dwMinorVersion * 100);
}

#ifndef LVS_EX_AUTOCHECKSELECT
#define LVS_EX_AUTOCHECKSELECT  0x08000000
#endif

void VSTEffectPriorityChange(HWND hList, int src, int dst)
{
	{
		VSTProcessingWaitLocker lk(g_processingLock);
		
		PluginData tmp;
		int tmp_index;
		int size = PluginList.size();

		mutex.lock();
		tmp = PluginList[dst];
		tmp_index= PluginList[dst].index;
		PluginList[dst] = PluginList[src];
		PluginList[src] = tmp;
		mutex.unlock();

		for (int i = 0; i < size; ++i) {
			updatePluginDataPointer(&PluginList[i]);
			PluginList[i].index = i; // id č\z
		}
	}

	SwapVSTItem(hList, dst, src);
}

void VSTEffectPriorityDown(HWND hwnd)
{
	HWND hList = GetDlgItem(hwnd, IDC_LIST1);
	int n = ListView_GetNextItem(hList, -1, LVNI_ALL | LVNI_SELECTED);
	if (n < 0 || n >= ListView_GetItemCount(hList) - 1)return;
	VSTEffectPriorityChange(hList, n, n + 1);
}

void VSTEffectPriorityUp(HWND hwnd)
{
	HWND hList = GetDlgItem(hwnd, IDC_LIST1);
	int n = ListView_GetNextItem(hList, -1, LVNI_ALL | LVNI_SELECTED);
	if (n <= 0)return;
	VSTEffectPriorityChange(hList, n, n - 1);
}

LRESULT VSTWindowProcInit(HWND hwnd, WPARAM wParam, LPARAM lParam, bool &bInitFlagPointer)
{
	HWND hList = GetDlgItem(hwnd, IDC_LIST1);

	if (GetOSVerInt() >= 6100) {
		ListView_SetExtendedListViewStyle(hList, LVS_EX_CHECKBOXES | LVS_EX_FULLROWSELECT);
	}else{
		ListView_SetExtendedListViewStyle(hList, LVS_EX_CHECKBOXES | LVS_EX_AUTOCHECKSELECT | LVS_EX_FULLROWSELECT);
	}

	LV_COLUMN lvc;
	lvc.mask		= LVCF_FMT | LVCF_WIDTH | LVCF_TEXT | LVCF_SUBITEM;
	lvc.fmt			= LVCFMT_LEFT;
	lvc.cx			= 200;
	lvc.pszText		= "VST Name";
	lvc.iSubItem	= 0;
	ListView_InsertColumn(hList, 0, &lvc);


	for (int i = 0; i < (int)PluginList.size(); i++) {
		InsertVSTItem(hList, &PluginList[i], i);
	}

	CreateChannelControl(hwnd);

	bInitFlagPointer = false;

	return TRUE;
}

LRESULT VSTWindowProcDropFiles(HWND hwnd, WPARAM wParam, LPARAM lParam)
{
	int old_size = PluginList.size();

	HDROP hDrop = (HDROP)wParam;
	int n = DragQueryFile(hDrop, 0xFFFFFFFF, NULL, 0);
	for(int i = 0; i < n; i++) {
		char fn[MAX_PATH + 4];
		DragQueryFile(hDrop, i, fn, MAX_PATH);
		
		extern bool VST_Loader( PluginData *pld, char *path, void (* PluginListBackDeleter)());
		extern void PluginList_DeleteEndBlock();
		
		// add
		mutex.lock();
		PluginList.push_back(PluginData());
		PluginData *pld = &PluginList[ PluginList.size() - 1 ];
		mutex.unlock();

		pld->bProsessing = false;

		VST_Loader(pld, fn, PluginList_DeleteEndBlock);
	}
	DragFinish(hDrop);

	// update
	for (int i = 0; i < old_size; ++i) {
		updatePluginDataPointer(&PluginList[i]);
	}

	HWND hList = GetDlgItem(hwnd, IDC_LIST1);
	for (int i = old_size; i < (int)PluginList.size(); ++i) {
		openEffectEditor(&PluginList[i], NULL, true);
		InsertVSTItem(hList, &PluginList[i], i);
	}

	return FALSE;
}

LRESULT VSTWindowProcNotify(HWND hwnd, WPARAM wParam, LPARAM lParam)
{
	NMHDR *pnmhdr = (NMHDR *)lParam;

	if (LOWORD(wParam) == IDC_LIST1){
		switch (pnmhdr->code) {
		case NM_DBLCLK:{	
			int idx = GetListIndex(GetDlgItem(hwnd, IDC_LIST1));
			if (idx == -1)return FALSE;
			openEffectEditor(&PluginList[idx], NULL);
			return FALSE;
		}
		case LVN_ITEMCHANGED: {
			LPNMLISTVIEW pnmlist = (LPNMLISTVIEW)lParam;
			if (pnmlist->iItem == -1)return FALSE;
			int imgmask;
			if ((imgmask = pnmlist->uNewState & LVIS_STATEIMAGEMASK) != (pnmlist->uOldState & LVIS_STATEIMAGEMASK)) {

				if (imgmask == INDEXTOSTATEIMAGEMASK(2)) {
					PluginList[pnmlist->iItem].bProsessing = true;
				}else if (imgmask == INDEXTOSTATEIMAGEMASK(1)){
					PluginList[pnmlist->iItem].bProsessing = false;
					PluginList[pnmlist->iItem].bClearVstBuffer = true;
				}
			}
			return FALSE;
		}
		default:
			return FALSE;
		}
	}

	return FALSE;
}

LRESULT VSTWindowProcCommand(HWND hwnd, WPARAM wParam, LPARAM lParam)
{
	WORD ControlId = LOWORD(wParam);
	switch (ControlId) {
	case IDC_VW_VSTADD: {
		g_VstLoaderIndex = -1; // global
		POINT pt;
		GetCursorPos(&pt);
		TrackPopupMenu( g_hVstItemMenu, TPM_LEFTALIGN, pt.x, pt.y, 0, hwnd, NULL );
		return FALSE;
	}
	case IDC_VW_VSTDEL:
	{
		VSTProcessingWaitLocker lk(g_processingLock);
		int n = -1;
		HWND hList = GetDlgItem(hwnd, IDC_LIST1);
		n = ListView_GetNextItem(hList, -1, LVNI_ALL | LVNI_SELECTED);
		if (n == -1)break;

		ListView_DeleteItem(hList, n);
		mutex.lock();
		PluginList.erase( PluginList.begin() + n );
		for (int i = 0; i < (int)PluginList.size(); ++i) {
			updatePluginDataPointer_unsafe(&PluginList[i]);
		}
		mutex.unlock();
		return FALSE;
	}
	case IDC_VW_VSTUP:
		VSTEffectPriorityUp(hwnd);
		return FALSE;
	case IDC_VW_VSTDOWN:
		VSTEffectPriorityDown(hwnd);
		return FALSE;
	case IDC_VW_VST_CONF:
		extern void OpenVSTPref(HWND hwnd);
		OpenVSTPref(hwnd);
		return FALSE;
	default:
		ChannelManager_Command(hwnd, ControlId, lParam);
		return FALSE;
	}
	return FALSE;
}

LRESULT CALLBACK VSTWindowProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	static bool bInit = true;
	switch (msg) {
	case WM_INITDIALOG:
		return VSTWindowProcInit(hwnd, wParam, lParam, bInit);
	case WM_DROPFILES:
		return VSTWindowProcDropFiles(hwnd, wParam, lParam);
	case WM_NOTIFY:
		if (bInit)break;
		return VSTWindowProcNotify(hwnd, wParam, lParam);
	case WM_COMMAND:
		return VSTWindowProcCommand(hwnd, wParam, lParam);
	case WM_CLOSE:
		bInit = true;
		hwndManager = NULL;
		DestroyWindow(hwnd);

		return FALSE;
	}

	return 0;
}

void __stdcall openVSTManager(HWND hwnd)
{
	if(hwndManager == NULL) {
		hwndManager = CreateDialog(hDllInst, MAKEINTRESOURCE(IDD_VSTWND), NULL, (DLGPROC)VSTWindowProc);
		ShowWindow(hwndManager, SW_SHOW);
	}else{
		SetFocus(hwndManager);
	}
}

// -----------------------------------------------------------

#define CUSTOMEDIT_SLIDER_MAX 2000

void CreateCustomEditorControl(HWND hwnd, int param_num)
{
	HWND hBaseLabel = GetDlgItem(hwnd, IDC_CUSTOMEDIT_LBL);
	DWORD dwLabelStyle = GetWindowLong(hBaseLabel, GWL_STYLE);
	RECT stLabelRect = GetWindowPosition(hBaseLabel);
	HFONT hFont = (HFONT)SendMessage(hwnd, WM_GETFONT, 0, 0);

	HWND hBaseSlider = GetDlgItem(hwnd, IDC_CUSTOMEDIT_SB1);
	DWORD dwSliderStyle = GetWindowLong(hBaseSlider, GWL_STYLE);
	RECT stSliderRect = GetWindowPosition(hBaseSlider);
	ScrollBar_SetRange(hBaseSlider, 0, CUSTOMEDIT_SLIDER_MAX, TRUE);
	ScrollBar_SetPos(hBaseSlider, 0, TRUE);

	HWND hBaseParam = GetDlgItem(hwnd, IDC_CUSTOMEDIT_LBP);
	DWORD dwParamStyle = GetWindowLong(hBaseParam, GWL_STYLE);
	RECT stParamRect = GetWindowPosition(hBaseParam);

	HWND childctrl;
	for (int k = 1, set_i = 0; k < param_num; ++k) {
		childctrl = CreateWindow("STATIC","VST", dwLabelStyle,
			stLabelRect.left, 
			k * 20 + stLabelRect.top, 
			stLabelRect.right - stLabelRect.left, 
			stLabelRect.bottom - stLabelRect.top, 
			hwnd, 
			(HMENU)(IDC_CUSTOMEDIT_LBL+(k)), hDllInst, NULL);
		SendMessage(childctrl, WM_SETFONT, (WPARAM)hFont, 1);

		childctrl = CreateWindow("STATIC","VST", dwParamStyle,
			stParamRect.left, 
			k * 20 + stParamRect.top, 
			stParamRect.right - stParamRect.left, 
			stParamRect.bottom - stParamRect.top, 
			hwnd, 
			(HMENU)(IDC_CUSTOMEDIT_LBP+(k)), hDllInst, NULL);
		SendMessage(childctrl, WM_SETFONT, (WPARAM)hFont, 1);

		childctrl = CreateWindow("SCROLLBAR","VST", dwSliderStyle,
			stSliderRect.left, 
			k * 20 + stSliderRect.top, 
			stSliderRect.right - stSliderRect.left, 
			stSliderRect.bottom - stSliderRect.top, 
			hwnd, 
			(HMENU)(IDC_CUSTOMEDIT_SB1+(k)), hDllInst, NULL);
		ScrollBar_SetRange(childctrl, 0, CUSTOMEDIT_SLIDER_MAX, TRUE);
		ScrollBar_SetPos(childctrl, 0, TRUE);
	}
}

void SetCustomEditorLabelText(HWND hwnd, int index, const char *str)
{
	SetWindowText(GetDlgItem(hwnd, IDC_CUSTOMEDIT_LBL + index), str);
}

void SetCustomEditorParamText(HWND hwnd, int index, const char *str)
{
	SetWindowText(GetDlgItem(hwnd, IDC_CUSTOMEDIT_LBP + index), str);
}

void SetCustomEditorSlider(HWND hwnd, int index, float n)
{
	ScrollBar_SetPos(GetDlgItem(hwnd, IDC_CUSTOMEDIT_SB1 + index), (long)(n * (float)CUSTOMEDIT_SLIDER_MAX), TRUE);
}

void GetCustomEditorLabelText(HWND hwnd, int index, char *str, size_t sz)
{
	GetWindowText(GetDlgItem(hwnd, IDC_CUSTOMEDIT_LBL + index), str, sz);
}

void GetCustomEditorParamText(HWND hwnd, int index, char *str, size_t sz)
{
	GetWindowText(GetDlgItem(hwnd, IDC_CUSTOMEDIT_LBP + index), str, sz);
}

float GetCustomEditorSlider(HWND hwnd, int index)
{
	return (float)ScrollBar_GetPos(GetDlgItem(hwnd, IDC_CUSTOMEDIT_SB1 + index)) / (float)CUSTOMEDIT_SLIDER_MAX;
}

void UpdateCustomEditorSlider(HWND hwnd, AEffect *effect)
{
	for (int i = 0; i < effect->numParams; ++i) {
		SetCustomEditorSlider(hwnd, i, effect->getParameter(effect, i));
		char str[256], str2[256];
		effect->dispatcher(effect, effGetParamName, i, 0, str, 0);
		SetCustomEditorLabelText(hwnd, i, str);

		effect->dispatcher(effect, effGetParamLabel, i, 0, str2, 0);
		effect->dispatcher(effect, effGetParamDisplay, i, 0, str, 0);
		strcat(str, str2);
		SetCustomEditorParamText(hwnd, i, str);
	}
}

LRESULT CALLBACK CustomEditCtrlProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch (msg) {
		case WM_INITDIALOG: {
			AEffect *eff = reinterpret_cast<AEffect *>(lParam);
			SetProp(hwnd, "__CUSTOMEDIT_VSTDATA__", (HANDLE)lParam);
			CreateCustomEditorControl(hwnd, eff->numParams);

			UpdateCustomEditorSlider(hwnd, eff);

			return TRUE;
		}

		case WM_COMMAND:
			switch(LOWORD(wParam)){
			case IDM_CUSTOMEDIT_UPDATEBAR:
				UpdateCustomEditorSlider(hwnd,  (AEffect *)GetProp(hwnd, "__CUSTOMEDIT_VSTDATA__") );
				break;
			}
			break;
		case WM_RBUTTONUP:{
			POINT pt;
			GetCursorPos(&pt);
			HMENU hMenu = LoadMenu(hDllInst, MAKEINTRESOURCE(IDR_MENU1));
			TrackPopupMenu(GetSubMenu(hMenu,0), TPM_LEFTALIGN | TPM_TOPALIGN ,pt.x, pt.y, 0, hwnd, NULL);
			DestroyMenu(hMenu);
			break;
		}
		case WM_HSCROLL:
		{
			UINT id = GetDlgCtrlID((HWND)lParam);
			long n = ScrollBar_GetPos((HWND)lParam);
			if (id >= IDC_CUSTOMEDIT_SB1 && id < IDC_CUSTOMEDIT_SB1+100) {
				int index = id - IDC_CUSTOMEDIT_SB1;
				switch(LOWORD(wParam)) {
				case SB_PAGELEFT:
					n -= 50;
				case SB_LINELEFT:
					n = max(0, n - 1);
					break;
				case SB_PAGERIGHT:
					n += 50;
				case SB_LINERIGHT:
					n = min(CUSTOMEDIT_SLIDER_MAX, n + 1);
					break;
				case SB_LEFT:
					n = 0;
					break;
				case SB_RIGHT:
					n = CUSTOMEDIT_SLIDER_MAX;
					break;
				case SB_THUMBPOSITION:
				case SB_THUMBTRACK:
					n = HIWORD(wParam);
					break;
				}

				ScrollBar_SetPos((HWND)lParam, n, TRUE);

				AEffect *eff = (AEffect *)GetProp(hwnd, "__CUSTOMEDIT_VSTDATA__");

				eff->setParameter(eff, index, GetCustomEditorSlider(hwnd, index));
				char str[256], str2[256];

				eff->dispatcher(eff, effGetParamLabel, index, 0, str2, 0);
				eff->dispatcher(eff, effGetParamDisplay, index, 0, str, 0);
				strcat(str, str2);
				SetCustomEditorParamText(hwnd, index, str);
			}
			break;
		}

		case WM_NOTIFY:
			break;
		case WM_DESTROY:
			RemoveProp(hwnd, "__CUSTOMEDIT_VSTDATA__");
			break;
		case WM_CLOSE:
			break;
	}
	return 0;
}

HWND CreateCustomEditorDialog(HWND hwnd, PluginData *pd)
{
	HWND chwnd = CreateDialogParam(hDllInst, MAKEINTRESOURCE(IDD_CUSTOMEDIT), hwnd, (DLGPROC)CustomEditCtrlProc, (LPARAM)pd->effect);
	ShowWindow(chwnd, SW_SHOW);
	return chwnd;
}



LRESULT CALLBACK SetVSTSettingDlgProc(HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{
	switch (msg){
		case WM_INITDIALOG:
			SetDlgItemText(hwnd, IDC_VSTS_VSTPATH, g_VstFolderPath);
			return TRUE;
		case WM_COMMAND:
			switch(LOWORD(wp)){
			case IDOK:
			{
				GetDlgItemText(hwnd, IDC_VSTS_VSTPATH, g_VstFolderPath, MAX_PATH);

				IniMemFile ini;
				ini.LoadFromFile("soundfont.ini");
				ini.WriteString("VST", "Path", g_VstFolderPath);
				ini.WriteInt("VST", "MaxChannel", g_MaxChannelVstCount);
				ini.SaveToFile();

				CreateVSTItemMenu();
				SendMessage(hwnd,WM_CLOSE,0,0);
				return true;
			}
			case IDCANCEL:
				SendMessage(hwnd,WM_CLOSE,0,0);
				return false;
			case IDC_VSTS_BTN_VSTPATH:
			{
				CMyFolderDialog fd(hwnd);
				fd.setTitle("VST Folder");
				if (fd.Execute( g_VstFolderPath )) {
					strcpy(g_VstFolderPath, fd.getFolder());
					SetDlgItemText(hwnd, IDC_VSTS_VSTPATH,g_VstFolderPath);
				}
				break;
			}
			default:
				break;
			}
			break;
		case WM_CLOSE:
			EndDialog(hwnd,true);
			break;
	}
	return FALSE;
}

void OpenVSTPref(HWND hwnd)
{
	DialogBoxA(hDllInst,MAKEINTRESOURCE(IDD_SETVSTSETTING),hwnd,(DLGPROC)SetVSTSettingDlgProc);
}